==> oscars_mgmnt_clms-0.0.8.dev20240410+42-py3-none-any.whl: oscars_mgmnt_clms/parser/netcdf.py <==
import os
from datetime import datetime as dt
from typing import Tuple

from netCDF4 import Dataset
import shapely
import numpy as np
import re
from oscars_mgmnt_clms.file_object import ObjectInfo
from oscars_mgmnt_clms.parser.base import BaseParser
from oscars_mgmnt_clms.parser.filename import parse_fname_date, FilenameParser


class CLMSGlobalParser_netCDF(FilenameParser):
    """Parser used to build GeoJSON metadata from CLMS Global Land products in netCDF format"""

    @property
    def filetype(self) -> str:
        return "netcdf"

    def __init__(self):
        super().__init__(template="CLMS_GLOBAL.template")

    def __decodePlatforms__(self, platformstr):
        import re
        # decodes the netCDF :platform attribute into a list of one or more
        # platforms, with consistent formatting

        # examples:
        # Proba-V => [PROBA-V]
        # SPOT 4 => [SPOT-4]
        # METOP A+B => [METOP-A, METOP-B]
        # Sentinel-1 A+B => [SENTINEL-1A, SENTINEL-1B]
        # METOP A+B, Sentinel-1 A+B => [METOP-A, METOP-B, SENTINEL-1A, SENTINEL-1B]
        # SENTINEL-3A AND SENTINEL-3B => [SENTINEL-3A, SENTINEL-3B]
        # SENTINEL3A AND SENTINEL3B => [SENTINEL-3A, SENTINEL-3B]
        # SENTINEL 3 (A AND B) => [SENTINEL-3A, SENTINEL-3B]
        # TERRA (EOS AM-1) => [TERRA]
        # N/A => [N/A]

        platformstr = (
            platformstr.upper()
            .replace("_", "-")
            .replace("SUOMI-NPP", "SUOMI NPP")
            .strip()
        )

        # support for platforms with LIE 250m with old name for TERRA/AQUA
        # TERRA (EOS AM-1) => [TERRA]
        for modis in ["AQUA", "TERRA"]:
            if modis in platformstr and " (" in platformstr:
                platformstr = platformstr.split(" (")[0]

        split_comma = platformstr.split(",")
        platforms = [x.strip() for x in split_comma]

        # Support for "AND" between platform names or platform serials
        # e.g. "SENTINEL-3A AND SENTINEL-3B" or "SENTINEL 3 (A AND B)"
        bracket_serials_regex = "\(([A-Z]) AND ([A-Z])\)"
        if " AND " in platformstr:
            repl_platforms = []
            for platf in platforms:
                regex_match_part = re.search(bracket_serials_regex, platf)
                if regex_match_part:
                    (platformbase, platformserial) = platf.split(" (")
                    platformbase = platformbase.replace(" ", "-")
                    sep = "-"
                    if "-" in platformbase:
                        sep = ""
                    repl_platforms.extend(
                        [
                            platformbase + sep + serial
                            for serial in regex_match_part.groups()
                        ]
                    )
                else:
                    repl_platforms.extend(platf.split(" AND "))
            platforms = repl_platforms

        # Support for SSM/SWI multi-mission products with eg "METOP B+C"
        if "+" in platformstr:
            repl_platforms = []
            for platf in platforms:
                if "+" in platf:
                    (platformbase, platformserial) = platf.split(" ")
                    split_serial = platformserial.split("+")
                    sep = "-"
                    if "-" in platformbase:
                        sep = ""
                    repl_platforms.extend(
                        [platformbase + sep + serial for serial in split_serial]
                    )
                else:
                    repl_platforms.extend(platf)
            platforms = repl_platforms

        # We covered the AND and + syntaxes, all remaining spaces are between platforms
        # with exception of "SUOMI NPP"
        repl_platforms = []
        for platf in platforms:
            if " " in platf and not "SUOMI NPP" == platf:
                repl_platforms.extend(platf.split(" "))
            else:
                repl_platforms.append(platf)
            platforms = repl_platforms

        # Support for LST, eg MSG4 -> MSG-4
        geos = ["GOES", "MSG", "HIMAWARI", "MTSAT"]
        lst_regex = "(MSG|MTSAT|HIMAWARI|GOES)([0-9]+|E|R)"
        if any([x in platformstr for x in geos]):
            repl_platforms = []
            for platf in platforms:
                regex_match = re.match(lst_regex, platf)
                if regex_match:
                    repl_platforms.append("-".join(regex_match.groups()))
                else:
                    repl_platforms.append(platf)
            platforms = repl_platforms

        # Hyphenate SENTINELs, eg SENTINEL3A => SENTINEL-3A
        sentinel_regex = "(SENTINEL)([0-6][A-Z]+)"
        if "SENTINEL" in platformstr:
            repl_platforms = []
            for platf in platforms:
                if "SENTINEL" in platf:
                    regex_match = re.match(sentinel_regex, platf)
                    if regex_match:
                        repl_platforms.append("-".join(regex_match.groups()))
                    else:
                        repl_platforms.append(platf)
                else:
                    repl_platforms.append(platf)
            platforms = repl_platforms

        return platforms

    def specific_treatment(self, product_file_info: ObjectInfo, values: dict):
        self.fix_products(product_file_info)

        # examples of supported filenames:
        #    c_gls_NDVI300_202204110000_GLOBE_OLCI_V2.0.1.nc
        #    c_gls_BA300-NRT_201901010000_GLOBE_S3_V3.1.1.nc
        fn = os.path.abspath(product_file_info.get_product_path())
        file_name = os.path.basename(fn).rpartition(".")[0]
        file_name_parts = file_name.split("_")

        ds = Dataset(fn)

        values["id"] = file_name
        values["title"] = ds.title

        (product_type, separator, product_group) = file_name_parts[2].partition("-")
        product_group = product_group.replace("STATIC-", "")
        if product_group:
            values["productgroupid"] = product_group

        values["producttype"] = product_type
        version = file_name_parts[-1]
        values["productVersion"] = version

        processingCenter = "VITO"
        if "institution" in ds.ncattrs():
            processingCenter = ds.institution.split(",")[0].upper()
        elif "processing_centre" in ds.ncattrs():
            processingCenter = ds.processing_centre.split(",")[0].upper()
        values["processingCenter"] = (
            processingCenter.replace(" NV", "").replace(" GMBH", "").replace(" ", "_")
        )

        # Coordinate rounding to 10 decimals
        # Allow one more digit for LST to avoid ES map parser error
        rounding = 11 if "land_surface_temperature" in fn else 10
        bbox = self.get_bbox(ds, rounding, product_group)
        values["westbbox"], values["southbbox"], values["eastbbox"], values["northbbox"] = bbox
        values["geometry"] = self.get_geometry(bbox)

        # LWQ refers to GLOBAL, rename to GLOBE for consistency with other products
        values["tileid"] = file_name_parts[-3].replace("GLOBAL", "GLOBE")

        if "time_coverage_start" in ds.ncattrs():
            values["beginposition"] = ds.time_coverage_start
            values["endposition"] = ds.time_coverage_end
        else:
            values["beginposition"] = values["nominaldatetime"]
            values["endposition"] = values["nominaldatetime"]

        # collection_dir can have 12.5km, 5km, 1km, 300m, 100m or no resolution info
        collection_dir = re.match(f"clms_global_([\\w.-]+)_{self.filetype}", values["parentidentifier"]).group(1)
        resolution_regex = re.compile("([\\d.]+)(m|km)")
        regex_result = resolution_regex.match(collection_dir.split("_")[1])

        if regex_result and product_group not in ["DGG", "TS"]:
            regex_result = resolution_regex.match(collection_dir.split("_")[1])
            factor = 1000 if regex_result.group(2)[0] == "k" else 1
            values["resolution"] = int(float(regex_result.group(1)) * factor)
        else:
            values["resolution"] = None

        values["platforms"] = self.__decodePlatforms__(ds.platform)

        # TODO extend for SWI DGG product
        values["conformsto"] = "https://www.opengis.net/def/crs/EPSG/0/4326"

        ds.close()

        return values["id"]

    def get_bbox(self, ds: Dataset, rounding: int, product_group: str) -> Tuple[float, float, float, float]:
        # netCDF stores center of pixel coordinates, OSCARS stores pixel top left coordinates
        # so we need to adjust with half a pixel
        westcenter = np.min(ds.variables["lon"])
        southcenter = np.min(ds.variables["lat"])
        eastcenter = np.max(ds.variables["lon"])
        northcenter = np.max(ds.variables["lat"])

        if "TS" == product_group:
            # SWI Time series: DGG grid is irregular in lat/lon degrees, no half pixel adjustment
            westbbox = np.round(westcenter, rounding)
            southbbox = np.round(southcenter, rounding)
            eastbbox = np.round(eastcenter, rounding)
            northbbox = np.round(northcenter, rounding)
        else:
            rasterxsize = len(ds.variables["lon"])
            rasterysize = len(ds.variables["lat"])

            step = (eastcenter - westcenter) / (rasterxsize - 1)
            # improve final decimals for SWI and PROBA-V/Sentinel-3 vegetation products
            for realstep in [0.1, 1.0 / 112, 1.0 / 336]:
                precision = 10
                if np.round(realstep, precision) == np.round(step, precision):
                    step = realstep
                    eastcenter = westcenter + ((rasterxsize - 1) * step)
                    southcenter = northcenter - ((rasterysize - 1) * step)
            halfpixel = step / 2.0

            # shift half pixel towards from center to top-left of pixel
            westbbox = np.round(westcenter - halfpixel, rounding)
            southbbox = np.round(southcenter - halfpixel, rounding)
            eastbbox = np.round(eastcenter + halfpixel, rounding)
            northbbox = np.round(northcenter + halfpixel, rounding)

        return westbbox, southbbox, eastbbox, northbbox

    def get_geometry(self, bbox: Tuple[float, float, float, float]) -> str:
        geometry = shapely.box(*bbox)
        geometry = shapely.make_valid(geometry)
        return shapely.to_geojson(geometry)


    @staticmethod
    def fix_products(fileobject):
        import os
        import re

        for info in list(fileobject.related):
            if info.name.endswith(".nc"):
                title = os.path.basename(info.name).split("_")[2].split("-")[0]
                re_match = re.search("\d", title)
                if re_match:
                    # if needed, use all characters up to first digit (eg NDVI300 -> NDVI)
                    title = title[0 : re_match.start()]

                info.title = title
                info.type = "application/netcdf"
                fileobject.related.remove(info)
                fileobject.datas.append(info)

==> oscars_mgmnt_clms-0.1.2.dev20240429+53-py3-none-any.whl: oscars_mgmnt_clms/parser/netcdf.py <==
import os
from typing import Tuple

from netCDF4 import Dataset
import numpy as np
import re
from oscars_mgmnt_clms.file_object import ObjectInfo
from oscars_mgmnt_clms.parser.filename import (
    FilenameParser,
    get_geometry,
    decode_platforms,
    parse_date,
    format_date,
    fix_bbox,
)


class CLMSGlobalParser_netCDF(FilenameParser):
    """Parser used to build GeoJSON metadata from CLMS Global Land products in netCDF format"""

    @property
    def filetype(self) -> str:
        return "netcdf"

    def __init__(self):
        super().__init__(template="CLMS_GLOBAL.template")

    def specific_treatment(self, product_file_info: ObjectInfo, values: dict):
        self.fix_products(product_file_info)

        # examples of supported filenames:
        #    c_gls_NDVI300_202204110000_GLOBE_OLCI_V2.0.1.nc
        #    c_gls_BA300-NRT_201901010000_GLOBE_S3_V3.1.1.nc
        fn = os.path.abspath(product_file_info.get_product_path())
        file_name = os.path.basename(fn).rpartition(".")[0]
        file_name_parts = file_name.split("_")

        ds = Dataset(fn)

        values["id"] = file_name
        values["title"] = ds.title

        (product_type, separator, product_group) = file_name_parts[2].partition("-")
        product_group = product_group.replace("STATIC-", "")
        if product_group:
            values["productgroupid"] = product_group

        values["producttype"] = product_type
        version = file_name_parts[-1]
        values["productVersion"] = version

        processingCenter = "VITO"
        if "institution" in ds.ncattrs():
            processingCenter = ds.institution.split(",")[0].upper()
        elif "processing_centre" in ds.ncattrs():
            processingCenter = ds.processing_centre.split(",")[0].upper()
        values["processingCenter"] = (
            processingCenter.replace(" NV", "").replace(" GMBH", "").replace(" ", "_")
        )

        # Coordinate rounding to 10 decimals
        # Allow one more digit for LST to avoid ES map parser error
        rounding = 11 if "land_surface_temperature" in fn else 10
        bbox = self.get_bbox(ds, rounding)
        (
            values["westbbox"],
            values["southbbox"],
            values["eastbbox"],
            values["northbbox"],
        ) = bbox
        values["geometry"] = get_geometry(bbox)

        # LWQ refers to GLOBAL, rename to GLOBE for consistency with other products
        values["tileid"] = file_name_parts[-3].replace("GLOBAL", "GLOBE")

        if "time_coverage_start" in ds.ncattrs():
            values["beginposition"] = format_date(parse_date(ds.time_coverage_start))
            values["endposition"] = format_date(parse_date(ds.time_coverage_end))
        else:
            values["beginposition"] = values["nominaldatetime"]
            values["endposition"] = values["nominaldatetime"]

        # collection_dir can have 12.5km, 5km, 1km, 300m, 100m or no resolution info
        collection_dir = re.match(
            f"clms_global_([\\w.-]+)_{self.filetype}", values["parentidentifier"]
        ).group(1)
        resolution_regex = re.compile("([\\d.]+)(m|km)")
        regex_result = resolution_regex.match(collection_dir.split("_")[1])
        if regex_result and product_group not in ["DGG", "TS"]:
            factor = 1000 if regex_result.group(2)[0] == "k" else 1
            values["resolution"] = int(float(regex_result.group(1)) * factor)
        else:
            values["resolution"] = None

        values["platforms"] = decode_platforms(ds.platform)

        # TODO extend for SWI DGG product
        values["conformsto"] = "https://www.opengis.net/def/crs/EPSG/0/4326"

        ds.close()

        return values["id"]

    def get_bbox(self, ds: Dataset, rounding: int) -> Tuple[float, float, float, float]:
        # netCDF stores center of pixel coordinates, OSCARS stores pixel top left coordinates
        # so we need to adjust with half a pixel
        westcenter = np.min(ds.variables["lon"])
        southcenter = np.min(ds.variables["lat"])
        eastcenter = np.max(ds.variables["lon"])
        northcenter = np.max(ds.variables["lat"])

        # SWI Time series: DGG grid is irregular in lat/lon degrees, no half pixel adjustment
        rasterxsize = ds.dimensions["lon"].size if "lon" in ds.dimensions else None
        rasterysize = ds.dimensions["lat"].size if "lat" in ds.dimensions else None

        return fix_bbox(
            (westcenter, southcenter, eastcenter, northcenter),
            rounding,
            rasterxsize,
            rasterysize,
        )

    @property
    def mimetype(self) -> str:
        return "application/netcdf"
